#include <xc.h>
#include "delay.h"
#include "temperature.h"
#include "display.h"

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover (Internal/External Switchover mode is disabled)
#pragma config FCMEN = ON       // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is enabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config VCAPEN = OFF     // Voltage Regulator Capacitor Enable bit (VCAP pin function disabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)

void init() {

    OSCCONbits.IRCF = 0x0F; /* raise internal oscillator clock to 16 Mhz        */
    while (OSCSTATbits.HFIOFR==0) {
        delay_1tcy();       /* wait until internal oscillator is ready          */
    }
    while (OSCSTATbits.HFIOFS==0) {
        delay_1tcy();       /* wait until oscillator stabilizes, within 0.5%    */
    }
    ANSELA = 0b00000001;    /* enable analogue input on PORT A, RA0 (v check)   */
    ANSELB = 0;             /* disable analogue operations on the other ports   */
    ANSELC = 0;
    ANSELD = 0;
    ANSELE = 0;
    TRISA  = 0b00000001;    /* set RA0 to input mode, for ADC operation    */
    TRISB  = 0;             /* set PORT B,C,D,E to output                       */
    TRISC  = 0;
    TRISD  = 0;

    FVRCONbits.ADFVR = 3;   /* Fixed Voltage Reference: 4.096v , 4mV per bit    */
    FVRCONbits.FVREN = 1;   /* Enable voltage reference                         */

/*  FVRRDY is always ‘1’ on PIC16F151X only. So commenting this.
 *   while (FVRCONbits.FVRRDY==0) {
 *       delay_1tcy();       // wait for the voltage reference to stabilize
 *   }
 */

    delay_ms(10); /* wait for voltage reference to stabilize, perhaps not needed */
    ADCON0bits.CHS = 0;     /* AN0 , battery voltage check */
    ADCON1bits.ADCS = 7;    /* 7 = internal oscillator                          */
    ADCON1bits.ADFM = 1;    /* [......xx] [xxxxxxxx] (right justified)          */
    ADCON1bits.ADPREF = 3;  /* voltage reference set to internal reference      */
    ADCON0bits.ADON = 1;    /* enable adc, leave it running all time*/
}

/* switches voltage reference between the three levels: 1.024, 2.048 , 4.096 
 * Note that the ADC only works with 2.048v or 4.096v settings on internal vref
 * Other options for ADC are external vref (1.8v or greater) or Vdd 
 */

void vref_set(char value) {
    if (FVRCONbits.ADFVR==value) return;
    if (value>3) return;
    FVRCONbits.FVREN = 0;
    FVRCONbits.ADFVR = value;
    FVRCONbits.FVREN = 1;
    /* on PIC16F1519 voltage reference is always "ready" due to the internal voltage
     * regulator the chip has, so in theory there's only switching between levels
     */
    while (FVRCONbits.FVRRDY==0) {
         /* wait for the voltage reference to stabilize   */
    }
    /* should  wait a bit just to be sure the reference stabilizes if needed
     * but this function will be called before displaying the temperature on led
     * display for some time, so the reference will have time to stabilize before
     * the next adc_read call.
     */
    /* delay_1ms();  */
    return;
}

unsigned short adc_read(char id) {
//    while (ADCON0bits.GO == 1) {
//   /* wait until previous conversion finishes, in case somehow this function
//      is called while adc is already doing a conversion
//    */
//    }
    ADCON0bits.CHS = id;    /* set the channel on the A/D converter             */
    //ADCON0bits.ADON = 1;  /* turn ON the ADC                                */
    delay_25us();           /* wait for sampling capacitor to charge            */
    delay_25us();
    ADCON0bits.ADGO = 1;    /* start the ADC process                            */
    while (ADCON0bits.ADGO == 1) {
                            /* wait until conversion is done                    */
    }
    //ADCON0bits.ADON = 0;    /* turn off ADC (not necessary but doesn't hurt     */
    unsigned short nr = (ADRESHbits.ADRESH << 8);
    nr = nr + ADRESLbits.ADRESL;
    return nr;

}

/* Check the input voltage. Show battery low if voltage below a threshold
 *
 * AD595 runs directly from 9V battery, because input voltage has to be about
 * 2v above the maximum temperature output (we want at least 4v output from chip,
 * which reports about 400C)
 * The MCU needs more than 4v to have a working 4v voltage reference, so a 5v
 * linear regulator will be used. Depending on what regulator is chosen, a
 * minimum of 5.5-7v is required (for example 1117 needs 1.2v, 7805 needs 2v).
 *
 * Input voltage connected to PORTA:0 through a voltage divider (~ 4:1),
 * therefore the MCU will see about 10v / 4 = 2v for a new 9v battery.
 *
 */

void voltage_show(){
    unsigned short temp;            /* temporary variable for voltage digits    */
    /* set pin AN0 to analogue input */
    TRISAbits.TRISA0 = 1;
    ANSELAbits.ANSA0 = 1;
    unsigned short value = adc_read(ADC_VOLTAGE); /* read battery voltage       */
//    display_setnumber(value,0);
//    display_show(2000);
    /* 10bit ADC [0..1023], 4.096v ref => 4.0039mV per bit
     * volts before the ~x:1 divider, in mV.
     * VOLTAGE_RATIO (the voltage divider ratio) may be floating point, so there
     * may be some small (rounding) error when converting back to integer
     * but it's tolerable here (it's only a few mV of error)
    */
    TRISAbits.TRISA0 = 0; /* turn it off, we're done with it */
    ANSELAbits.ANSA0 = 0;

    float valuef = value * 4.00391 * VOLTAGE_RATIO;
    /* voltage is in mV.
     * If value is < 10v, then we basically have 4 digits already(ex. 6789 mV).
     * If more than 10v, need to divide by 10 to fit in our four digits display
     * Show decimal point and value so we always have just two decimals.
    */
    temp = (unsigned short) valuef;

    if (temp < 10000) {
        temp = temp / 10; /* remove last digit, ex. show 6.78v for 6789 mV */
    } else {
        temp = temp / 10; /* show 4 digits ex show 12.34 for 12345 mV    */
    }

    if (valuef <= VOLTAGE_THRESHOLD) {
        while (1) {     /* alternate between -Lo- and voltage until turned off  */
            display_setdigit(0,'-');
            display_setdigit(1,'L');
            display_setdigit(2,'o');
            display_setdigit(3,'-');
            display_show(2000);           /* show -Lo- for ~ 2000 ms          */
            display_setnumber(temp,1);    /* put voltage on the four digits   */
            display_show(2000);           /* show voltage for ~ 2000 ms       */
        }
    } else {                /* just show the voltage once, for about 2 seconds  */
            display_setnumber(temp,1);    /* put voltage on the four digits   */
            display_show(2000);           /* show voltage for ~ 2000 ms       */
    }

}

void temperature_show() {
    unsigned short decimals = 0; /* how many decimals temperature has           */
    unsigned short temps = 0;   /* temperature without decimals                 */
    unsigned short value;
    float multiplier = 4.0391; /* 4096mV / 1023 steps */

    float voltage=0;
    /* make sure all other adc inputs are disabled, enable temperature input pin */
    TRISAbits.TRISA1 = 1;
    ANSELAbits.ANSA1 = 1;
    value = adc_read(ADC_TEMPERATURE); /* read  voltage  */
    /* multiplier constant is initialed for 4.096 vref. We may have lowered the
     * Vref for better accuracy so let's check and correct multiplier if needed
     */
    if (FVRCONbits.ADFVR==2) {
        multiplier = 2.001955;
        /* as long as we're here, if the result is close to 1023, then set the
         * Vref so next measurement will be done using the updated Vref
         */
        if (value > 1010) {
           vref_set(3); // set the Vref to 4.096v
        }
    } else {
        /* Vref is set to 4.096v and multiplier is OK, but if the adc value tells
         * us the input voltage is below ~ 1.8v, let's start using the 2.048v
         * on next measurements for better accuracy
         */
        if (value < 450) {
            vref_set(2); // set the Vref to 2.048v
        }
    }
    
    //value = value * 4;  /* 10bit ADC, 4.096v ref, 1023 values => 4.039mV per bit     */
    voltage = value * multiplier * TEMPERATURE_RATIO; /* real v, before the ~4:1 divider   */

    float tempf = temperature_from_voltage(voltage);
    if (TEMPERATURE_AVERAGE == 1 ) {
        /* averaging enabled, so pass it through the averaging function */
        tempf = temperature_average(tempf);
    }
    if (tempf < 100){
        decimals = 1;
        temps = (unsigned short) (tempf*100);
    } else {
        decimals = 2;
        temps = (unsigned short) (tempf*10);
    }
    display_setnumber(temps,decimals); /* put the temperature on the 4 digits */
    display_show(TEMPERATURE_DISPLAYTIME); /* show temperature for n ms */
}

void main() {
    init();
    voltage_show();
    while (1) {
        temperature_show();
    }
}

